import random
import torch
import collections

from drmax.agents.agent import Agent
from drmax.utils.memory import ReplayMemory
from drmax.models import QNetwork, get_optimizer
from drmax.utils.save import net_directory, get_filename


class DQN(Agent):

    def __init__(self, name, env_name, state_dimension, actions, gamma, learning_rate, epsilon, epsilon_decay,
                 replay_memory_size, minibatch_size, double_dqn, hidden_layers, target_update_period,
                 n_gradient_updates, load_constant_net, device, optimizer_name='adam',
                 momentum=0):
        Agent.__init__(self, name=name, actions=actions, gamma=gamma)

        self.env_name = env_name
        self.n_a = len(actions)  # Only discrete actions are implemented
        self.state_dimension = state_dimension
        self.learning_rate = learning_rate
        self.epsilon = epsilon
        self.epsilon_decay = epsilon_decay
        self.replay_memory_size = replay_memory_size
        self.minibatch_size = minibatch_size
        self.double_dqn = double_dqn
        self.hidden_layers = hidden_layers
        self.target_update_period = target_update_period
        self.n_gradient_updates = n_gradient_updates
        self.load_constant_net = load_constant_net
        self.device = device
        self.optimizer_name = optimizer_name
        self.momentum = momentum

        self.loss_function = torch.nn.MSELoss(reduction='mean')  # reduction: 'mean' (default), 'sum' or 'none'
        self.learns_Q_function = True  # Called for visualization by external methods
        self.has_replay_buffer = True  # Called for visualization by external methods

        # To be reset between instances:
        self.replay_memory = ReplayMemory(self.replay_memory_size)

        # Create Q-networks
        self.Q = QNetwork(input_size=self.state_dimension, output_size=self.n_a, hidden_layers=self.hidden_layers)
        self.Q_target = QNetwork(input_size=self.state_dimension, output_size=self.n_a,
                                 hidden_layers=self.hidden_layers)

        # Load existing Q-networks if required
        if self.load_constant_net:
            net_dir = net_directory(env_name)
            q_max = 1.0 / (1.0 - self.gamma)
            filename = get_filename(key='constant_net', net_dir=net_dir, hidden_layers=self.hidden_layers, cst=q_max)
            self.Q.load(filename=filename, device=self.device)
            self.Q_target.load(filename=filename, device=self.device)

        self.optimizer = get_optimizer(optimizer_name=self.optimizer_name, parameters=self.Q.parameters(),
                                       learning_rate=self.learning_rate, momentum=self.momentum)
        self.update_counter = 0
        self.losses = []
        self.prev_s = None
        self.prev_a = None

    def re_init(self):
        """
        Re-initialization for multiple instances.
        :return: None
        """
        self.__init__(name=self.name, env_name=self.env_name, state_dimension=self.state_dimension,
                      actions=self.actions, gamma=self.gamma, learning_rate=self.learning_rate, epsilon=self.epsilon,
                      epsilon_decay=self.epsilon_decay, replay_memory_size=self.replay_memory_size,
                      minibatch_size=self.minibatch_size, double_dqn=self.double_dqn, hidden_layers=self.hidden_layers,
                      target_update_period=self.target_update_period, n_gradient_updates=self.n_gradient_updates,
                      load_constant_net=self.load_constant_net, device=self.device, optimizer_name=self.optimizer_name,
                      momentum=self.momentum)

    def act(self, s, r, is_terminal):
        """
        Acting method called online during learning.
        :param s: current state of the agent
        :param r: (float) received reward for the previous transition
        :param is_terminal: (bool) True if s is terminal
        :return: return the epsilon-greedy action wrt the current learned Q-function.
        """
        self.update(self.prev_s, self.prev_a, r, s, is_terminal)

        a = self.epsilon_greedy_action(s)

        self.prev_a = a
        self.prev_s = s

        if self.epsilon_decay is not None:
            self.epsilon *= self.epsilon_decay

        return a

    def get_batch(self):
        transitions = self.replay_memory.sample(self.minibatch_size)
        batch = list(zip(*transitions))  # Create 5 lists of states, actions, rewards, next_states, terminal

        # Convert to tensors
        states = torch.FloatTensor(batch[0]).to(self.device)
        '''
        # TODO for continuous action space
        if self.is_action_space_continuous:
            actions = torch.FloatTensor(batch[1]).to(self.device)
        else:
            actions = torch.LongTensor(batch[1]).to(self.device)
        '''
        actions = torch.LongTensor(batch[1]).to(self.device)
        rewards = torch.FloatTensor(batch[2]).unsqueeze(1).to(self.device)
        next_states = torch.FloatTensor(batch[3]).to(self.device)
        is_terminal = torch.FloatTensor(batch[4]).unsqueeze(1).to(self.device)

        return states, actions, rewards, next_states, is_terminal

    def update_target(self):
        """
        Update the target network by copying the weights of the Q network.
        :return: None
        """
        for qt_parameters, q_parameters in zip(self.Q.parameters(), self.Q_target.parameters()):
            # Exponential decay
            # target_param.data.copy_((1-tau)*target_param.data + tau*nn_param.data)
            q_parameters.data.copy_(qt_parameters.data)

    def update(self, s, a, r, s_p, is_terminal):
        """
        Update the current Q-function
        :param s: state
        :param a: action
        :param r: (float) reward
        :param s_p: next state
        :param is_terminal: (bool) True if s_p is terminal
        :return: None
        """
        # 1) Push new sample in the replay memory
        if s is not None and a is not None:
            self.replay_memory.push(s, a, r, s_p, is_terminal)

        # 2) Update the model if enough samples in the replay memory to produce a minibatch
        if len(self.replay_memory) >= self.minibatch_size:
            self.update_counter += 1
            for _ in range(self.n_gradient_updates):
                states, actions, rewards, next_states, done = self.get_batch()

                # Compute Q(s, a): compute first Q(s), then select the columns of the taken actions
                q_s_a = self.Q(states).gather(1, actions.unsqueeze(1))

                # Compute Q(s_p)
                if self.double_dqn:
                    # Compute a_p = argmax_a Q(s_{t+1}, a)
                    next_actions = self.Q(next_states).argmax(1).unsqueeze(1)
                    # Compute Q_target(s_{t+1}, a_p) for a double DQN
                    q_sp = self.Q_target(next_states).gather(1, next_actions)
                else:
                    # Compute Q(s_p) and select the maximum
                    q_sp = self.Q_target(next_states).max(1)[0].unsqueeze(1)

                # Compute the expected Q values : y[i]= r[i] + gamma * Q(s[i+1], a[i+1])
                y_target = rewards + (1 - done) * self.gamma * q_sp
                y_target = y_target.detach()

                # Compute loss
                loss = self.loss_function(q_s_a, y_target)

                # Zero gradients, perform a backward pass (gradient update), and update the weights.
                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()

                # Target update
                if self.update_counter == self.target_update_period:
                    self.update_target()
                    self.update_counter = 0

                # Record loss
                self.losses.append(loss.item())

    def end_of_episode(self):
        """
        Reset between episodes within the same MDP.
        :return: None
        """
        self.prev_s = None
        self.prev_a = None

    def epsilon_greedy_action(self, s):
        if random.random() < self.epsilon:
            return random.choice(self.actions)
        else:
            return self.greedy_action(s)

    def greedy_action(self, s):
        """
        Compute the greedy action wrt the Q-network
        :param s: state at which the upper-bound is evaluated
        :param f: input function of (s, a)
        :return: return the greedy action and the achieved value by this action.
        """
        with torch.no_grad():
            # state = torch.FloatTensor(s).to(self.device)
            return self.actions[self.Q(torch.FloatTensor(s).to(self.device)).cpu().detach().argmax().item()]

    def q_forward(self, s, a):
        """
        Evaluate Q function at (s, a)
        :param s: state
        :param a: action
        :return: Q-value of (s, a)
        """
        s = torch.FloatTensor([s]).to(self.device)
        a = torch.LongTensor([a]).to(self.device)
        return self.Q(s).gather(1, a.unsqueeze(1)).item()

    def get_replay_buffer(self):
        """
        Return a list of the states in the replay memory.
        If index is provided, only save a subset of the list items
        :param index:
        :return:
        """
        states = [(s[0].x, s[0].y) for s in self.replay_memory.memory]
        return dict(collections.Counter(states))

    def get_backup(self):
        return {'loss': self.losses}



"""
Pendulum system
Inspired from https://github.com/openai/gym/blob/master/gym/envs/classic_control/pendulum.py
"""

import numpy as np

from drmax.environments.environment import Environment
from drmax.environments.state import State


class PendulumState(State):
    def __init__(self, theta, theta_dot):
        State.__init__(self, data=[theta, theta_dot])
        self.theta = theta
        self.theta_dot = theta_dot

    def dim(self):
        return 2

    def raw(self):
        return np.array([self.theta, self.theta_dot])


class Pendulum(Environment):

    def __init__(self, name, gamma, sparse_reward):

        self.sparse_reward = sparse_reward
        self.max_speed = 8.0
        self.max_torque = 2.0
        step = 1.0
        self.available_torque = np.arange(-self.max_torque, self.max_torque + step, step)
        actions = np.arange(len(self.available_torque))  # Discrete actions
        self.dt = 0.05
        self.gravity = 9.81
        self.mass = 1.0
        self.length = 1.0
        self.high = np.array([np.pi, self.max_speed], dtype=np.float32)

        self.max_cost = self.cost(np.pi, self.max_speed, self.max_torque)

        Environment.__init__(self, name=name, actions=actions, gamma=gamma)

    def get_state_dimension(self):
        return 2

    def get_state_dtype(self):
        return [float, float]

    def get_state_magnitude(self):
        return -self.high, self.high

    def get_initial_state(self):
        return PendulumState(
            theta=3.14,
            theta_dot=0.0
        )

    def step(self, s, a):
        theta, theta_dot = s.theta, s.theta_dot
        torque = self.available_torque[a]

        theta_dot_p = theta_dot + (- 3.0 * self.gravity / (2.0 * self.length) * np.sin(theta + np.pi) +
                                   3.0 / (self.mass * self.length ** 2) * torque) * self.dt
        theta_dot_p = np.clip(theta_dot_p, -self.max_speed, self.max_speed)
        theta_p = wrap_angle(theta + theta_dot_p * self.dt)

        s_p = PendulumState(theta=theta_p, theta_dot=theta_dot_p)

        cost_ratio = self.cost(theta_p, theta_dot_p, torque) / self.max_cost
        if self.sparse_reward:
            r = 1.0 if cost_ratio < 0.01 else 0.0
        else:
            r = 1.0 - cost_ratio

        return r, s_p, self.is_terminal(s_p)

    def is_terminal(self, s):
        return False

    def cost(self, theta, theta_dot, torque):
        return theta ** 2 + 0.1 * theta_dot ** 2 + 0.001 * torque ** 2

    def get_info(self):
        """
        Get general information to be saved on disk.
        """
        return {
            'name': self.name,
            'actions': self.actions,
            'gamma': self.gamma,
            'max_speed': self.max_speed,
            'max_torque': self.max_torque,
            'dt': self.dt,
            'gravity': self.gravity,
            'mass': self.mass,
            'length': self.length
        }


def wrap_angle(x):
    """
    Wrap angle in [- pi, pi]
    :param x: (float) angle
    :return: (float) wrapped angle
    """
    return ((x + np.pi) % (2 * np.pi)) - np.pi



import time
import pandas
import numpy as np
import dill
import multiprocessing
import torch
from tqdm import trange

import drmax.utils.configreader as reader
import drmax.utils.save as save
import drmax.utils.plot as plot
from drmax.environments.gridworld.gridworld import GridWorld
from drmax.models import QNetwork


def run_dill_encoded(payload):
    fun, args = dill.loads(payload)
    return fun(*args)


def apply_async(pool, fun, args):
    payload = dill.dumps((fun, args))
    return pool.apply_async(run_dill_encoded, (payload,))


def rl_experiment(agent, env, instance_number, n_instances, n_episodes, n_steps, timeout, save_ma, verbose=True):
    gamma = env.gamma

    # Save buffers
    trajectory = [[0, 0.0, 0.0, 0.0, 0.0]]  # Rewards trajectory
    trajectory_buffer = [trajectory[-1][1:]]

    steps_per_episode = np.zeros(n_episodes, dtype=int)
    return_per_episode = np.zeros(n_episodes)
    discounted_return_per_episode = np.zeros(n_episodes)
    time_per_episode = np.zeros(n_episodes)

    if verbose:
        print('Environment :', env.name)
        print('Agent       :', agent.name)
        print('Instance    :', str(instance_number), "/", str(n_instances))

    # For each episode.
    step_counter = 1
    is_timeout = False
    for episode in trange(n_episodes, desc='Episodes'):

        # Compute initial state and reward
        s = env.get_initial_state()
        r = 0.0
        is_terminal = False
        init_time = time.clock()
        reached_terminal_state = False

        for step in range(n_steps):

            # Compute the agent's policy
            a = agent.act(s, r, is_terminal)

            # Execute in MDP
            r, s_p, is_terminal = env.step(s, a)

            # Save trajectory
            if save_ma == 1:
                trajectory.append([
                    step_counter,  # step
                    r,  # reward
                    r * gamma ** float(step),  # discounted reward
                    r + trajectory[-1][3],  # cumulative reward
                    r * gamma ** float(step) + trajectory[-1][4]  # cumulative discounted reward
                ])
            else:
                trajectory_buffer.append([
                    r,  # reward
                    r * gamma ** float(step),  # discounted reward
                    r + trajectory_buffer[-1][2],  # cumulative reward
                    r * gamma ** float(step) + trajectory_buffer[-1][3]  # cumulative discounted reward
                ])
                if len(trajectory_buffer) == save_ma + 1:
                    trajectory_buffer = trajectory_buffer[1:]
                    trajectory.append([step_counter] + list(np.mean(trajectory_buffer, axis=0)))
                    trajectory_buffer = [trajectory_buffer[-1]]

            # Store return per episode
            return_per_episode[episode] += r
            discounted_return_per_episode[episode] += r * (gamma ** step)

            # Terminal / timeout check
            is_timeout = True if timeout is not None and step_counter == timeout else False
            if is_terminal or is_timeout:
                steps_per_episode[episode] = step + 1
                reached_terminal_state = True

                # Save remaining trajectory_buffer before break if timeout or end of last episode
                if save_ma > 1:
                    if is_timeout or episode + 1 == n_episodes:
                        trajectory_buffer = trajectory_buffer[1:]
                        trajectory.append([step_counter] + list(np.mean(trajectory_buffer, axis=0)))

                break

            # Update pointer
            s = s_p
            step_counter += 1

        # Record time and number of steps
        if not reached_terminal_state:
            steps_per_episode[episode] = n_steps
        time_per_episode[episode] = time.clock() - init_time

        # A final update
        agent.act(s, r, is_terminal)

        # Reset the MDP, tell the agent the episode is over
        # env.reset()
        agent.end_of_episode()

        # Timeout break
        if is_timeout:
            break

    # Re-initialization for multiple instances
    ag_backup = agent.get_backup()

    # If Gridworld and agent is Q-based, save Q-values for visual
    if type(env) == GridWorld:
        ag_backup = grid_world_visuals(env, agent, ag_backup)

    # Re-initialization for multiple instances
    agent.re_init()

    trajectory.pop(0)  # cleaning trick
    trajectory_df = pandas.DataFrame(trajectory, columns=['step', 'r', 'dr', 'cr', 'cdr'])

    if verbose:
        print('Successfully ended', env.name, agent.name, '(instance', str(instance_number), "/", str(n_instances), ')')

    return trajectory_df, steps_per_episode, return_per_episode, discounted_return_per_episode, time_per_episode, \
           ag_backup


def grid_world_visuals(env, agent, ag_backup):
    """
    Backup for visuals in gridworld environment.
    :param env:
    :param agent:
    :param ag_backup:
    :return:
    """
    if ag_backup is None:  # Replace by dict if None
        ag_backup = {}

    try:
        if agent.learns_Q_function:
            ag_backup['q_map'] = env.get_q_map(agent)
    except AttributeError:
        pass

    try:
        if agent.has_replay_buffer:
            ag_backup['replay_buffer'] = agent.get_replay_buffer()
    except AttributeError:
        pass

    return ag_backup


def run(ag, env, instance_number, n_ins, n_ep, n_st, timeout, save_ma, verbose, do_save, ag_env_dir):
    tr, steps_per_episode, return_per_episode, discounted_return_per_episode, time_per_episode, ag_backup = \
        rl_experiment(agent=ag, env=env, instance_number=instance_number, n_instances=n_ins, n_episodes=n_ep,
                      n_steps=n_st, timeout=timeout, save_ma=save_ma, verbose=verbose)

    # Save this environment-agent pair results
    if do_save:
        save.benchmark_save(ag_env_dir=ag_env_dir, agent=ag, env=env, n_episodes=n_ep, n_steps=n_st,
                            instance_number=instance_number, trajectory=tr, steps_per_episode=steps_per_episode,
                            return_per_episode=return_per_episode,
                            discounted_return_per_episode=discounted_return_per_episode,
                            time_per_episode=time_per_episode, ag_backup=ag_backup)


def benchmark(config, do_run=True, do_save=True, exp_dir=None, do_plot=True, do_show=False, verbose=True):
    # Create saving directory and save configuration
    if do_save:
        if exp_dir is None:
            exp_dir = save.datetime_directory(verbose=verbose)
        config_filename = save.get_filename(key='global_config', exp_dir=exp_dir)
        save.dict_to_csv(dictionary=config, filename=config_filename)  # CSV
        save.save_obj(obj=config, filename=config_filename)  # PKL

    # Read configuration file
    env_ag, n_env, n_episodes, n_steps, n_instances, timeouts, n_processes, is_multi_proc, save_ma, env_names =\
        reader.benchmark_settings_from_config(config=config, exp_dir=exp_dir)

    # Check values
    assert n_env == len(n_episodes), 'Please provide a number of episodes for each environment (provided ' \
                                     + str(len(n_episodes)) + ')'
    assert n_env == len(n_steps), 'Please provide a number of steps for each environment (provided ' \
                                  + str(len(n_steps)) + ')'
    assert n_env == len(n_instances), 'Please provide a number of instances for each environment (provided ' \
                                      + str(len(n_instances)) + ')'
    assert n_env == len(timeouts), 'Please provide a timeout value for each environment (provided '\
                                   + str(len(timeouts)) + ')'
    assert len(env_names) == len(set(env_names)), 'Please provide unique names for environments: ' + str(env_names)

    # Create pool of workers
    if is_multi_proc:
        pool = multiprocessing.Pool(processes=n_processes)
        jobs = []

    # Verbose
    if verbose:
        print('Running: ', n_env, 'environments with (respectfully)', [len(env_ag[env]) for env in env_ag],
              'agent(s) and', n_instances, 'instance(s).')

    if do_run:
        for env in env_ag:
            # 0) Save environment info
            if do_save:
                env_config_filename = save.get_filename(key='env_config', exp_dir=exp_dir, env_name=env.name)
                save.dict_to_csv(dictionary=env.get_info(), filename=env_config_filename)  # CSV
                save.save_obj(obj=env.get_info(), filename=env_config_filename)  # PKL

            # 1) Run all agents for this environment
            n_ep = n_episodes.pop(0)
            n_st = n_steps.pop(0)
            n_ins = n_instances.pop(0)
            timeout = timeouts.pop(0)
            for ag in env_ag[env]:
                if do_save:
                    ag_env_dir = save.ag_env_directory(exp_dir, ag, env)
                    save.create_dir_at(ag_env_dir)

                for instance_number in range(1, n_ins + 1):
                    if is_multi_proc:
                        job = apply_async(pool, run, (ag, env, instance_number, n_ins, n_ep, n_st, timeout,
                                                      save_ma, verbose, do_save, ag_env_dir))
                        jobs.append(job)
                    else:
                        run(ag, env, instance_number, n_ins, n_ep, n_st, timeout, save_ma, verbose, do_save, ag_env_dir)

    # Achieve all jobs
    if is_multi_proc:
        for job in jobs:
            job.get()

    # Verbose
    if verbose:
        print('End of run: ', n_env, 'environments with (respectfully)', [len(env_ag[env]) for env in env_ag],
              'agent(s) and', n_instances, 'instance(s).')

    # Plot all agents for this environment
    if do_plot:
        for env in env_ag:
            plot.benchmark_plot(exp_dir=exp_dir, agents=env_ag[env], environment=env,
                                markers_freq=config['markers_freq'], reward_plot_ma=config['reward_plot_ma'],
                                return_plot_ma=config['return_plot_ma'], do_show=do_show)


def generate_uniform_tensors(n_tensors, dtype, dimension, magnitude):
    """
    TODO: this function must be in utils AND should be called in ODQN.
    :param n_tensors: (int)
    :param dtype: ()
    :param dimension: (int)
    :param magnitude: (tuple)
    :return: (tensor)
    """
    assert len(set(dtype)) == 1, 'Multiple data type not implemented yet.'
    dtype = dtype[0]
    min_magnitude, max_magnitude = magnitude
    uniform_scale = torch.diag(torch.tensor(max_magnitude - min_magnitude, dtype=torch.float))
    uniform_add = torch.tensor(min_magnitude, dtype=torch.float)
    tensors = torch.rand(size=(n_tensors, dimension))
    tensors = torch.add(uniform_add, torch.matmul(tensors, uniform_scale))
    if dtype is float:
        return tensors
    elif dtype is int:
        return torch.floor(tensors)
    else:
        raise ValueError('Data type [' + dtype + '] not implemented in uniform data generation.')


def train_constant_net(config, env_config, do_save):
    # Create environment
    env = reader.environment_from_env_config(env_config)

    # Retrieve environment dimensions
    state_dimension = env.get_state_dimension()
    state_dtype = env.get_state_dtype()
    state_magnitude = env.get_state_magnitude()
    n_a = len(env.actions)

    # Initlialize Q-network
    net = QNetwork(input_size=state_dimension, output_size=n_a, hidden_layers=config['hidden_layers'])

    # Create uniform input
    batch = generate_uniform_tensors(n_tensors=config['batch_size'], dtype=state_dtype, dimension=state_dimension,
                                     magnitude=state_magnitude)

    # Train constant Q-network
    cst = 1.0 / (1.0 - env_config['parameters']['gamma'])
    norm = 0.5 / float(config['minibatch_size'])
    optimizer = torch.optim.Adam(net.parameters(), lr=config['learning_rate'])
    for _ in trange(config['n_epochs'], desc='Epochs'):
        # Sample minibatch
        indice = torch.randint(low=0, high=len(batch), size=(config['minibatch_size'],))
        mini_batch = batch[indice]

        # Forward + compute loss
        y = net(mini_batch)
        loss = norm * torch.add(input=y, other=-cst).pow(2).sum()

        # Zero gradients, perform a backward pass (gradient update), and update the weights.
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

    # Save constant Q-network
    if do_save:
        net_dir = save.net_directory(env_name=env_config['parameters']['name'])
        save.create_dir_at(net_dir, verbose=False)
        filename = save.get_filename(key='constant_net', net_dir=net_dir, hidden_layers=config['hidden_layers'],
                                     cst=cst)
        net.save(filename)


def train_multiple_constant_nets(config, do_save=True, verbose=True):
    """
    Fit a Q-network to 1 / (1 - gamma).
    :param config: (dict)
    :param do_save: (bool)
    :param verbose: (bool)
    :return: None
    """

    # Create pool of workers
    is_multi_proc, n_processes = reader.multiprocessing_settings_from_config(config=config)
    if is_multi_proc:
        pool = multiprocessing.Pool(processes=n_processes)
        jobs = []

    # Run all trainings
    for env_config in config['environments']:
        if is_multi_proc:
            job = apply_async(pool, train_constant_net, (config, env_config, do_save))
            jobs.append(job)
        else:
            train_constant_net(config=config, env_config=env_config, do_save=do_save)

    # Achieve all jobs
    if is_multi_proc:
        for job in jobs:
            job.get()



"""
Acrobot system
Inspired from https://github.com/openai/gym/blob/master/gym/envs/classic_control/acrobot.py
"""

import numpy as np

from drmax.environments.environment import Environment
from drmax.environments.state import State


class AcrobotState(State):
    def __init__(self, theta_1, theta_2, theta_dot_1, theta_dot_2):
        State.__init__(self, data=[theta_1, theta_2, theta_dot_1, theta_dot_2])
        self.theta_1 = theta_1
        self.theta_2 = theta_2
        self.theta_dot_1 = theta_dot_1
        self.theta_dot_2 = theta_dot_2

    def dim(self):
        return 4

    def raw(self):
        return np.array([self.theta_1, self.theta_2, self.theta_dot_1, self.theta_dot_2])


class Acrobot(Environment):

    def __init__(self, name, gamma, torque_noise_max=0.0):
        Environment.__init__(self, name=name, actions=[0, 1, 2], gamma=gamma)

        self.dt = 0.2
        self.link_length_1 = 1.0  # [m]
        self.link_length_2 = 1.0  # [m]
        self.link_mass_1 = 1.0  # [kg] mass of link 1
        self.link_mass_2 = 1.0  # [kg] mass of link 2
        self.link_com_pos_1 = 0.5  #: [m] position of the center of mass of link 1
        self.link_com_pos_2 = 0.5  #: [m] position of the center of mass of link 2
        self.link_moi = 1.  # moments of inertia for both links
        self.max_velocity_1 = 4.0 * np.pi
        self.max_velocity_2 = 9.0 * np.pi
        self.available_torque = [-1.0, 0.0, +1.0]
        self.torque_noise_max = torque_noise_max

        self.book_or_nips = 'book'  # use dynamics equations from the nips paper or the book

        self.high = np.array([np.pi, np.pi, self.max_velocity_1, self.max_velocity_2], dtype=np.float32)

    def get_state_dimension(self):
        return 4

    def get_state_dtype(self):
        return [float, float, float, float]

    def get_state_magnitude(self):
        return -self.high, self.high

    def get_initial_state(self):
        return AcrobotState(
            theta_1=0.0,
            theta_2=0.0,
            theta_dot_1=0.0,
            theta_dot_2=0.0
        )

    def step(self, s, a):
        """
        :param s: state
        :param a: action
        :return: r, s_p, is_terminal(s_p)
        """
        s = s.raw()
        torque = self.available_torque[a]
        if self.torque_noise_max > 0.0:
            torque += np.random.uniform(-self.torque_noise_max, self.torque_noise_max)
        s_augmented = np.append(s, torque)

        s_p = rk4(self._dsdt, s_augmented, [0, self.dt])
        s_p = s_p[-1]  # only get final time-step of integration
        s_p = s_p[:4]  # omit action

        s_p[0] = wrap_angle(s_p[0])
        s_p[1] = wrap_angle(s_p[1])
        s_p[2] = bound(s_p[2], -self.max_velocity_1, self.max_velocity_1)
        s_p[3] = bound(s_p[3], -self.max_velocity_2, self.max_velocity_2)

        s_p = AcrobotState(
            theta_1=s_p[0],
            theta_2=s_p[1],
            theta_dot_1=s_p[2],
            theta_dot_2=s_p[3]
        )

        terminal = self.is_terminal(s_p)
        r = 0.0 if not terminal else 1.0

        return r, s_p, terminal

    def is_terminal(self, s):
        return bool(-np.cos(s.theta_1) - np.cos(s.theta_2 + s.theta_1) > 1.)

    def _dsdt(self, s_augmented, t):
        m1 = self.link_mass_1
        m2 = self.link_mass_1
        l1 = self.link_length_1
        lc1 = self.link_com_pos_1
        lc2 = self.link_com_pos_1
        i1 = self.link_moi
        i2 = self.link_moi

        g = 9.8
        a = s_augmented[-1]
        s = s_augmented[:-1]
        theta1 = s[0]
        theta2 = s[1]
        d_theta1 = s[2]
        d_theta2 = s[3]
        d1 = m1 * lc1 ** 2 + m2 * \
            (l1 ** 2 + lc2 ** 2 + 2 * l1 * lc2 * np.cos(theta2)) + i1 + i2
        d2 = m2 * (lc2 ** 2 + l1 * lc2 * np.cos(theta2)) + i2
        phi2 = m2 * lc2 * g * np.cos(theta1 + theta2 - np.pi / 2.)
        phi1 = - m2 * l1 * lc2 * d_theta2 ** 2 * np.sin(theta2) \
               - 2 * m2 * l1 * lc2 * d_theta2 * d_theta1 * np.sin(theta2)  \
            + (m1 * lc1 + m2 * l1) * g * np.cos(theta1 - np.pi / 2) + phi2
        if self.book_or_nips == "nips":
            # the following line is consistent with the description in the paper
            dd_theta2 = (a + d2 / d1 * phi1 - phi2) / (m2 * lc2 ** 2 + i2 - d2 ** 2 / d1)
        else:
            # the following line is consistent with the java implementation and the book
            dd_theta2 = (a + d2 / d1 * phi1 - m2 * l1 * lc2 * d_theta1 ** 2 * np.sin(theta2) - phi2) \
                / (m2 * lc2 ** 2 + i2 - d2 ** 2 / d1)
        dd_theta1 = -(d2 * dd_theta2 + phi1) / d1
        return d_theta1, d_theta2, dd_theta1, dd_theta2, 0.0


def bound(x, x_min, x_max):
    """
    Truncates x between x_min and x_max
    :param x:
    :param x_min:
    :param x_max:
    :return:
    """
    return min(max(x, x_min), x_max)


def rk4(derivatives, y0, t, *args, **kwargs):
    """
    Integrate 1D or ND system of ODEs using 4-th order Runge-Kutta.
    This is a toy implementation which may be useful if you find
    yourself stranded on a system w/o scipy.  Otherwise use
    :func:`scipy.integrate`.

     Example 1 ::
        ## 2D system
        def derivatives(x,t):
            d1 =  x[0] + 2*x[1]
            d2 =  -3*x[0] + 4*x[1]
            return (d1, d2)
        dt = 0.0005
        t = arange(0.0, 2.0, dt)
        y0 = (1,2)
        yout = rk4(derivatives, y0, t)

    Example 2::
        ## 1D system
        alpha = 2
        def derivatives(x,t):
            return -alpha*x + exp(-t)
        y0 = 1
        yout = rk4(derivatives, y0, t)

    If you have access to scipy, you should probably be using the
    scipy.integrate tools rather than this function.

    :param derivatives: the derivative of the system and has the signature ``dy = derivs(yi, ti)``
    :param y0: initial state vector
    :param t: sample times
    :param args: additional arguments passed to the derivative function
    :param kwargs: additional keyword arguments passed to the derivative function
    :return: Runge-Kutta approximation of the ODE
    """
    try:
        n_y = len(y0)
    except TypeError:
        y_out = np.zeros((len(t),), np.float_)
    else:
        y_out = np.zeros((len(t), n_y), np.float_)
    y_out[0] = y0

    for i in np.arange(len(t) - 1):
        thist = t[i]
        dt = t[i + 1] - thist
        dt2 = dt / 2.0
        y0 = y_out[i]
        k1 = np.asarray(derivatives(y0, thist, *args, **kwargs))
        k2 = np.asarray(derivatives(y0 + dt2 * k1, thist + dt2, *args, **kwargs))
        k3 = np.asarray(derivatives(y0 + dt2 * k2, thist + dt2, *args, **kwargs))
        k4 = np.asarray(derivatives(y0 + dt * k3, thist + dt, *args, **kwargs))
        y_out[i + 1] = y0 + dt / 6.0 * (k1 + 2 * k2 + 2 * k3 + k4)
    return y_out


def wrap_angle(x):
    """
    Wrap angle in [- pi, pi]
    :param x: (float) angle
    :return: (float) wrapped angle
    """
    return ((x + np.pi) % (2 * np.pi)) - np.pi



import torch
import gym
import multiprocessing

from drmax.utils.save import agents_config_to_md

from drmax.agents.random import Random
from drmax.agents.rmax import RMax
from drmax.agents.qlearning import QLearning
from drmax.agents.dqn import DQN
from drmax.agents.odqn import ODQN

from drmax.environments.gridworld.gridworld import GridWorld
from drmax.environments.cartpole import CartPole
from drmax.environments.mountain_car import MountainCar
from drmax.environments.acrobot import Acrobot
from drmax.environments.pendulum import Pendulum
import drmax.environments.gym_binding as gym_binding


def expand(v, n):
    return n * [v] if type(v) is not list else v


def filtered_config_from_config(config):
    """
    Remove items from 'environments', 'n_episodes', 'n_steps', 'n_instances', 'timeout' lists in config.
    The removed items are those whose index matches a 0 in config['filter'].
    :param config: (dict) main configuration dict
    :return: (dict) the filtered main configuration dict
    """
    if 'filter' in config.keys() and config['filter'] is not None:
        f = config['filter']
        for key in ['environments', 'n_episodes', 'n_steps', 'n_instances', 'timeout']:
            if type(config[key]) is list:
                config[key] = [config[key][i] for i in range(len(f)) if f[i]]
    return config


def multiprocessing_settings_from_config(config):
    n_processes = multiprocessing.cpu_count() if config['n_processes'] is None else config['n_processes']
    is_multi_proc = n_processes > 1
    return is_multi_proc, n_processes


def benchmark_settings_from_config(config, exp_dir):
    env_ag = environments_agents_from_config(config, exp_dir)
    n_env = len(env_ag.keys())
    n_episodes = expand(config['n_episodes'], n_env)
    n_steps = expand(config['n_steps'], n_env)
    n_instances = expand(config['n_instances'], n_env)
    timeouts = expand(config['timeout'], n_env)
    is_multi_proc, n_processes = multiprocessing_settings_from_config(config=config)
    save_ma = config['save_ma']
    env_names = [ep['parameters']['name'] for ep in config['environments']]

    return env_ag, n_env, n_episodes, n_steps, n_instances, timeouts, n_processes, is_multi_proc, save_ma, env_names


def agent_from_class(agent_class, p, env, device):
    if agent_class == 'random':
        return Random(name=p['name'], actions=env.actions, gamma=env.gamma)
    elif agent_class == 'rmax':
        return RMax(name=p['name'], actions=env.actions, gamma=env.gamma, n_known=p['n_known'],
                    epsilon_q=p['epsilon_q'])
    elif agent_class == 'qlearning':
        return QLearning(name=p['name'], actions=env.actions, gamma=env.gamma, learning_rate=p['learning_rate'],
                         epsilon=p['epsilon'])
    elif agent_class == 'dqn':
        return DQN(name=p['name'], env_name=env.name, state_dimension=env.get_state_dimension(), actions=env.actions,
                   gamma=env.gamma, learning_rate=p['learning_rate'], epsilon=p['epsilon'],
                   epsilon_decay=p['epsilon_decay'], replay_memory_size=p['replay_memory_size'],
                   minibatch_size=p['minibatch_size'], double_dqn=p['double_dqn'], hidden_layers=p['hidden_layers'],
                   target_update_period=p['target_update_period'], n_gradient_updates=p['n_gradient_updates'],
                   load_constant_net=p['load_constant_net'], device=device)
    elif agent_class == 'odqn':
        return ODQN(name=p['name'], state_dimension=env.get_state_dimension(), actions=env.actions, gamma=env.gamma,
                    learning_rate=p['learning_rate'], replay_memory_size=p['replay_memory_size'],
                    minibatch_size=p['minibatch_size'], double_dqn=p['double_dqn'], hidden_layers=p['hidden_layers'],
                    target_update_period=p['target_update_period'], n_gradient_updates=p['n_gradient_updates'],
                    data_ratio=p['data_ratio'], alpha=p['alpha'], alpha_one_time=p['alpha_one_time'],
                    state_augmentation_type=p['state_augmentation_type'],
                    state_magnitude=env.get_state_magnitude(), state_dtype=env.get_state_dtype(),
                    cov_multiplier=p['cov_multiplier'], optimistic_factor=p['optimistic_factor'],
                    include_symmetric=p['include_symmetric'], device=device)
    else:
        raise ValueError('Agent class [' + agent_class + '] not found.')


def agents_from_config(config, env, exp_dir, param_in_name=False):
    agents = []
    device = device_from_config(config)

    for a in config['agents']:
        grid_search = a['grid_search']
        generic_name = a['parameters']['name']

        if grid_search:  # Grid search: build cartesian product of all parameters sets
            # Retrieve each parameter combination
            # del a['parameters']['name']
            p_grid = [{'name': generic_name}]
            for key, val_list in a['parameters'].items():
                if key is not 'name':
                    buffer = []
                    for i in range(len(p_grid)):
                        for j in range(len(val_list)):
                            p_grid_i = p_grid[i].copy()
                            p_grid_i[key] = val_list[j]
                            if param_in_name:
                                p_grid_i['name'] += '_' + key + '=' + str(p_grid_i[key])
                            buffer.append(p_grid_i)
                    p_grid = buffer

            # Rename agents by numbering them if parameters not included in name
            if not param_in_name:
                for i in range(len(p_grid)):
                    p_grid[i]['name'] += '_' + str(i + 1)

            # Save all configurations
            agents_config_to_md(exp_dir, parameters_list=p_grid)

            # Build agents
            for p in p_grid:
                agents.append(agent_from_class(agent_class=a['class'], p=p, env=env, device=device))

        else:  # Single agent
            p = a['parameters']
            agents_config_to_md(exp_dir, parameters_list=[p])
            agents.append(agent_from_class(agent_class=a['class'], p=p, env=env, device=device))

    return agents


def environment_from_env_config(env_config):
    env_class = env_config['class']
    p = env_config['parameters']

    if env_class == 'gridworld':
        return GridWorld(name=p['name'], gamma=p['gamma'], grid_name=p['grid_name'], size=p['size'],
                         slip_probability=p['slip_probability'], is_goal_terminal=p['is_goal_terminal'])
    elif env_class == 'cartpole':
        return CartPole(name=p['name'], gamma=p['gamma'], sparse_reward=p['sparse_reward'])
    elif env_class == 'mountaincar':
        return MountainCar(name=p['name'], gamma=p['gamma'], sparse_reward=p['sparse_reward'])
    elif env_class == 'acrobot':
        return Acrobot(name=p['name'], gamma=p['gamma'])
    elif env_class == 'pendulum':
        return Pendulum(name=p['name'], gamma=p['gamma'], sparse_reward=p['sparse_reward'])
    elif env_class == 'gym':
        return gym_binding.make(name=p['name'], gamma=p['gamma'])
    elif env_class == 'atari':
        # env = gym.make(p['name'])
        # env.frameskip = p['frameskip']
        return gym.make(p['name'])
    else:
        raise ValueError('Environment class [' + env_class + '] not implemented.')


def environments_agents_from_config(config, exp_dir):
    environments_agents = {}
    for env_config in config['environments']:
        env = environment_from_env_config(env_config)
        agents_config_to_md(exp_dir, env_name=env.name)
        environments_agents[env] = agents_from_config(config, env, exp_dir)

    return environments_agents


def device_from_config(config):
    if config['gpu'] and torch.cuda.is_available():
        return torch.device('cuda')
    else:
        return torch.device('cpu')



"""
Mountain car environment
"""

import math
import numpy as np

from drmax.environments.environment import Environment
from drmax.environments.state import State


class MountainCarState(State):
    def __init__(self, x, x_dot):
        State.__init__(self, data=[x, x_dot])
        self.x = x
        self.x_dot = x_dot

    def dim(self):
        return 2

    def raw(self):
        return np.array([self.x, self.x_dot])


class MountainCar(Environment):

    def __init__(self, name, gamma, sparse_reward):
        Environment.__init__(self, name=name, actions=[0, 1, 2], gamma=gamma)

        self.sparse_reward = sparse_reward
        self.min_position = -1.2
        self.max_position = 0.6
        self.max_speed = 0.07
        self.goal_position = 0.5
        self.goal_velocity = 0.5 * self.max_speed
        self.force = 0.001
        self.force *= 0.5  # Increased difficulty
        self.gravity = 0.0025

        self.low = np.array([self.min_position, -self.max_speed], dtype=np.float32)
        self.high = np.array([self.max_position, self.max_speed], dtype=np.float32)

    def get_state_dimension(self):
        return 2

    def get_state_dtype(self):
        return [float, float]

    def get_state_magnitude(self):
        return self.low, self.high

    def get_initial_state(self):
        return MountainCarState(0.0, 0.0)

    def step(self, s, a):
        """
        :param s: state
        :param a: action
        :return: r, s_p, is_terminal(s_p)
        """
        x, x_dot = s.x, s.x_dot

        x_dot += (a - 1) * self.force + math.cos(3 * x) * (-self.gravity)
        x_dot = np.clip(x_dot, -self.max_speed, self.max_speed)
        x += x_dot
        x = np.clip(x, self.min_position, self.max_position)
        if x <= self.min_position and x_dot < 0.0:
            x_dot = 0.0

        s_p = MountainCarState(x, x_dot)

        done = self.is_terminal(s_p)

        if self.sparse_reward:
            r = float(done)
        else:
            r = 0.99 * float(done) + 0.01 / (1.0 + abs(x - self.goal_position))

        return r, s_p, done

    def is_terminal(self, s):
        return bool(s.x >= self.goal_position and s.x_dot <= self.goal_velocity)

    def get_info(self):
        """
        Get general information to be saved on disk.
        """
        return {
            'name': self.name,
            'actions': self.actions,
            'gamma': self.gamma,
            'min_position': self.min_position,
            'max_position': self.max_position,
            'max_speed': self.max_speed,
            'goal_position': self.goal_position,
            'goal_velocity': self.goal_velocity,
            'force': self.force,
            'gravity': self.gravity
        }



from collections.abc import Sequence

import numpy as np


class State(Sequence):
    """
    Abstract State class
    """

    def __init__(self, data=[], is_terminal=False):
        self.data = data
        self._is_terminal = is_terminal

    def features(self):
        return np.array(self.data).flatten()

    def get_data(self):
        return self.data

    def get_num_feats(self):
        return len(self.features())

    def is_terminal(self):
        return self._is_terminal

    def set_terminal(self, is_term=True):
        self._is_terminal = is_term

    def __hash__(self):
        if type(self.data).__module__ == np.__name__:
            # Numpy arrays
            return hash(str(self.data))
        elif self.data.__hash__ is None:
            return hash(tuple(self.data))
        else:
            return hash(self.data)

    def __str__(self):
        return "s." + str(self.data)

    def __eq__(self, other):
        if isinstance(other, State):
            return self.data == other.data
        return False

    def __getitem__(self, index):
        return self.data[index]

    def __len__(self):
        return len(self.data)



import random
from collections import deque


class ReplayMemory:

    def __init__(self, capacity):
        """
        Simple replay memory structure storing a list of transitions (s, a, r, s_p, is_terminal)
        :param capacity: (int) capacity of the replay memory. If None, unbounded capacity.
        """
        self.capacity = capacity
        self.memory = []
        self.position = 0

    def push(self, *transition):
        if self.capacity is None:
            self.memory.append(transition)
        else:
            if len(self.memory) < self.capacity:
                self.memory.append(None)
            self.memory[self.position] = transition
            self.position = (self.position + 1) % self.capacity

    def sample(self, batch_size):
        return random.sample(self.memory, batch_size)

    def write(self, file_name):
        data = ''.join(map(sample_to_str, self.memory))
        with open(file_name, 'w') as file:
            file.write(data)

    def __len__(self):
        return len(self.memory)


class NStepsReplayMemory(ReplayMemory):

    def __init__(self, capacity, n_step, gamma):
        super().__init__(capacity)
        self.n_step = n_step
        self.gamma = gamma
        self.nstep_memory = deque()

    def _process_n_step_memory(self):
        s_mem, a_mem, R, si_, done = self.nstep_memory.popleft()
        if not done:
            for i in range(self.n_step-1):
                si, ai, ri, si_, done = self.nstep_memory[i]
                R += ri * self.gamma ** (i+1)
                if done:
                    break

        return [s_mem, a_mem, R, si_, done]

    def push(self, *transition):
        self.nstep_memory.append(transition)
        while len(self.nstep_memory) >= self.n_step or (self.nstep_memory and self.nstep_memory[-1][4]):
            nstep_transition = self._process_n_step_memory()
            super().push(*nstep_transition)


def sample_to_str(transition):
    s, a, r, s_, d = transition
    data = [list(s), list(a), r, list(s_), 1-int(d)]
    return ' ; '.join(map(str, data)) + '\n'



import torch


def get_optimizer(parameters, optimizer_name=None, learning_rate=0.05, momentum=0):
    if optimizer_name == 'adam':
        return torch.optim.Adam(parameters, lr=learning_rate)
    else:  # Default is SGD
        return torch.optim.SGD(parameters, lr=learning_rate, momentum=momentum)


class QNetwork(torch.nn.Module):

    def __init__(self, input_size, output_size, hidden_layers):
        super().__init__()

        self.layers = torch.nn.ModuleList([torch.nn.Linear(input_size, hidden_layers[0])])
        for i in range(1, len(hidden_layers)):
            self.layers.append(torch.nn.Linear(hidden_layers[i-1], hidden_layers[i]))
        self.output = torch.nn.Linear(hidden_layers[-1], output_size)

    def forward(self, x):
        for la in self.layers:
            x = torch.relu(la(x))
        return self.output(x)

    def save(self, filename):
        torch.save(self.state_dict(), filename)

    def load(self, filename, device):
        self.load_state_dict(torch.load(filename, map_location=device))



"""
Implementation of an R-Max agent [Brafman and Tennenholtz 2003]
Use Value Iteration to compute the R-Max upper-bound following [Strehl et al 2009].
"""

import random
import numpy as np
from collections import defaultdict

from drmax.agents.agent import Agent


def compute_n_model_samples_high_confidence(epsilon, delta, n_states):
    """
    Compute the number of model samples required for epsilon accurate model in L1-norm
    with probability at least 1 - delta.
    :param epsilon: (float)
    :param delta: (float)
    :param n_states: (int)
    :return: (int)
    """
    m_r_min = np.log(2. / delta) / (2. * epsilon ** 2.)
    m_t_min = 2. * (np.log(2. ** float(n_states) - 2.) - np.log(delta)) / (epsilon ** 2.)
    return int(max(m_r_min, m_t_min)) + 1


class RMax(Agent):
    """
    Implementation of an R-Max agent [Brafman and Tennenholtz 2003]
    Use Value Iteration to compute the R-Max upper-bound following [Strehl et al 2009].
    """

    def __init__(
            self,
            actions,
            gamma=0.9,
            r_max=1.0,
            v_max=None,
            n_known=None,
            epsilon_q=0.1,
            epsilon_m=0.1,
            delta=0.05,
            n_states=None,
            name="rmax"
    ):
        """
        :param actions: action space of the environment
        :param gamma: (float) discount factor
        :param r_max: (float) known upper-bound on the reward function
        :param v_max: (float) known upper-bound on the value function
            (if None, deduce v_max from r_max)
        :param n_known: (int) count after which a state-action pair is considered known
            (only set n_known if delta and epsilon are not defined)
            (if None, deduce n_known from (delta, n_states, epsilon_m))
        :param epsilon_q: (float) precision of value iteration algorithm for Q-value computation
        :param epsilon_m: (float) precision of the learned models in L1 norm
        :param delta: (float) models are learned epsilon_m-closely with probability at least 1 - delta
        :param n_states: (int) number of states
        :param name: (str)
        """
        Agent.__init__(self, name=name, actions=actions, gamma=gamma)
        self.r_max = r_max
        self.v_max = v_max
        if v_max is None:
            self.v_max = self.r_max / (1.0 - gamma)

        self.epsilon_q = epsilon_q
        self.epsilon_m = epsilon_m
        self.delta = delta
        self.n_states = n_states
        self.n_known = n_known
        if n_known is None:
            self.n_known = compute_n_model_samples_high_confidence(epsilon_m, delta, n_states)

        # Nb of value iterations
        self.vi_n_iter = int(np.log(1.0 / (epsilon_q * (1.0 - self.gamma))) / (1.0 - self.gamma))

        self.U, self.R, self.T, self.counter = self.empty_memory_structure()
        self.prev_s = None
        self.prev_a = None

        self.learns_Q_function = True

    def re_init(self):
        """
        Re-initialization for multiple instances.
        :return: None
        """
        self.__init__(actions=self.actions, gamma=self.gamma, r_max=self.r_max, v_max=self.v_max, n_known=self.n_known,
                      epsilon_q=self.epsilon_q, epsilon_m=self.epsilon_m, delta=self.delta, n_states=self.n_states,
                      name=self.name)

    def reset(self):
        """
        Reset the attributes to initial state (called between instances).
        :return: None
        """
        self.U, self.R, self.T, self.counter = self.empty_memory_structure()
        self.prev_s = None
        self.prev_a = None

    def act(self, s, r, is_terminal):
        """
        Acting method called online during learning.
        :param s: current state of the agent
        :param r: (float) received reward for the previous transition
        :param is_terminal: (bool) True if s is terminal
        :return: return the greedy action wrt the current learned model.
        """
        self.update(self.prev_s, self.prev_a, r, s)

        a = self.greedy_action(s, self.U)

        self.prev_a = a
        self.prev_s = s

        return a

    def end_of_episode(self):
        """
        Reset between episodes within the same MDP.
        :return: None
        """
        self.prev_s = None
        self.prev_a = None

    def empty_memory_structure(self):
        """
        Empty memory structure:
        U[s][a] (float): upper-bound on the Q-value
        R[s][a] (float): average reward
        T[s][a][s'] (float): probability of the transition
        counter[s][a] (int): number of times the state action pair has been sampled
        :return: U, R, T, counter
        """
        return defaultdict(lambda: defaultdict(lambda: self.v_max)),\
               defaultdict(lambda: defaultdict(float)),\
               defaultdict(lambda: defaultdict(lambda: defaultdict(float))),\
               defaultdict(lambda: defaultdict(int))

    def is_known(self, s, a):
        return self.counter[s][a] >= self.n_known

    def get_nb_known_sa(self):
        return sum([self.is_known(s, a) for s, a in self.counter.keys()])

    def update(self, s, a, r, s_p):
        """
        Updates transition and reward dictionaries with the input transition
        tuple if the corresponding state-action pair is not known enough.
        :param s: int state
        :param a: int action
        :param r: float reward
        :param s_p: int next state
        :return: None
        """
        if s is not None and a is not None:
            if self.counter[s][a] < self.n_known:
                self.counter[s][a] += 1
                normalizer = 1. / float(self.counter[s][a])

                self.R[s][a] = self.R[s][a] + normalizer * (r - self.R[s][a])
                self.T[s][a][s_p] = self.T[s][a][s_p] + normalizer * (1. - self.T[s][a][s_p])
                for _s_p in self.T[s][a]:
                    if _s_p not in [s_p]:
                        self.T[s][a][_s_p] = self.T[s][a][_s_p] * (1 - normalizer)

                if self.counter[s][a] == self.n_known:
                    self.update_upper_bound()

    def greedy_action(self, s, f):
        """
        Compute the greedy action wrt the input function of (s, a).
        :param s: state at which the upper-bound is evaluated
        :param f: input function of (s, a)
        :return: return the greedy action.
        """
        a_star = random.choice(self.actions)
        u_star = f[s][a_star]
        for a in self.actions:
            u_s_a = f[s][a]
            if u_s_a > u_star:
                u_star = u_s_a
                a_star = a
        return a_star

    def update_upper_bound(self):
        """
        Update the upper bound on the Q-value function.
        Called when a new state-action pair is known.
        :return: None
        """
        for _ in range(self.vi_n_iter):
            for s in self.R:
                for a in self.R[s]:
                    if self.is_known(s, a):
                        u_p = 0.
                        for s_p in self.T[s][a]:
                            u_p += max([self.U[s_p][a_p] for a_p in self.actions]) * self.T[s][a][s_p]
                        self.U[s][a] = self.R[s][a] + self.gamma * u_p

    def q_forward(self, s, a):
        """
        Evaluate Q function at (s, a)
        :param s: state
        :param a: action
        :return: Q-value of (s, a)
        """
        return self.U[s][a]



"""
Cart-pole system implemented by Rich Sutton et al.
Copied from http://incompleteideas.net/sutton/book/code/pole.c
permalink: https://perma.cc/C9ZM-652R
"""

import numpy as np

from drmax.environments.environment import Environment
from drmax.environments.state import State


class CartPoleState(State):
    def __init__(self, x, x_dot, theta, theta_dot):
        State.__init__(self, data=[x, x_dot, theta, theta_dot])
        self.x = x
        self.x_dot = x_dot
        self.theta = theta
        self.theta_dot = theta_dot

    def dim(self):
        return 4

    def raw(self):
        return np.array([self.x, self.x_dot, self.theta, self.theta_dot])


class CartPole(Environment):

    def __init__(self, name, gamma, sparse_reward):
        Environment.__init__(self, name=name, actions=[0, 1], gamma=gamma)

        self.sparse_reward = sparse_reward
        self.gravity = 9.8
        self.masscart = 1.0
        self.masspole = 0.1
        self.total_mass = (self.masspole + self.masscart)
        self.length = 0.5  # actually half the pole's length
        self.polemass_length = (self.masspole * self.length)
        self.force_mag = 10.0
        self.tau = 0.02  # seconds between state updates
        self.kinematics_integrator = 'euler'

        # Thresholds for terminal state
        self.theta_threshold_radians = 12.0 * np.pi / 180.0
        self.x_threshold = 2.4

        # Angle limit set to 2 * theta_threshold_radians so failing observation is still within bounds.
        # Empirical estimates of maximum: [4.8, 2.7, 0.21, 3]
        self.high = np.array([
            self.x_threshold,
            2.7,  # np.finfo(np.float32).max,
            self.theta_threshold_radians,
            3.0  # np.finfo(np.float32).max
        ], dtype=np.float32)

        self.steps_beyond_done = None

    def get_state_dimension(self):
        return 4

    def get_state_dtype(self):
        return [float, float, float, float]

    def get_state_magnitude(self):
        return -self.high, self.high

    def get_initial_state(self):
        return CartPoleState(0., 0., 0., 0.)

    def step(self, s, a):
        """
        :param s: state
        :param a: action
        :return: r, s_p, is_terminal(s_p)
        """
        x, x_dot, theta, theta_dot = s.raw()

        force = self.force_mag if a == 1 else -self.force_mag
        cos_theta = np.cos(theta)
        sin_theta = np.sin(theta)

        temp = (force + self.polemass_length * theta_dot ** 2 * sin_theta) / self.total_mass
        theta_acc = (self.gravity * sin_theta - cos_theta * temp) / (
                    self.length * (4.0 / 3.0 - self.masspole * cos_theta ** 2 / self.total_mass))
        x_acc = temp - self.polemass_length * theta_acc * cos_theta / self.total_mass

        if self.kinematics_integrator == 'euler':
            x = x + self.tau * x_dot
            x_dot = x_dot + self.tau * x_acc
            theta = theta + self.tau * theta_dot
            theta_dot = theta_dot + self.tau * theta_acc
        else:  # semi-implicit euler
            x_dot = x_dot + self.tau * x_acc
            x = x + self.tau * x_dot
            theta_dot = theta_dot + self.tau * theta_acc
            theta = theta + self.tau * theta_dot

        s_p = CartPoleState(x, x_dot, theta, theta_dot)

        done = self.is_terminal(s_p)

        if self.sparse_reward:
            r = 1.0 if abs(theta) < 0.05 * self.theta_threshold_radians and abs(s.x) < 0.33 * self.x_threshold else 0.0
        else:
            # r = 1.0 - abs(theta) / self.theta_threshold_radians
            r = 0.0 if done else 1.0

        return r, s_p, done

    def is_terminal(self, s):
        return bool(
            abs(s.x) > self.x_threshold
            or abs(s.theta) > self.theta_threshold_radians
        )

    def get_info(self):
        """
        Get general information to be saved on disk.
        """
        return {
            'name': self.name,
            'actions': self.actions,
            'gamma': self.gamma,
            'gravity': self.gravity,
            'masscart': self.masscart,
            'masspole': self.masspole,
            'total_mass': self.total_mass,
            'length': self.length,
            'polemass_length': self.polemass_length,
            'force_mag': self.force_mag,
            'tau': self.tau,
            'kinematics_integrator': self.kinematics_integrator,
            'theta_threshold_radians': self.theta_threshold_radians,
            'x_threshold': self.x_threshold
        }



import os
import csv
import pandas
import pickle
from datetime import datetime


def create_dir_at(path, verbose=True):
    try:
        os.mkdir(path)
    except OSError:
        if verbose:
            print("Creation of the directory %s failed" % path)
    else:
        if verbose:
            print("Successfully created the directory %s " % path)


def csv_ext(filename):
    return filename if filename[-4:] == '.csv' else filename + '.csv'


def pkl_ext(filename):
    return filename if filename[-4:] == '.pkl' else filename + '.pkl'


def save_obj(obj, filename):
    with open(pkl_ext(filename), 'wb') as f:
        pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)


def load_obj(filename):
    with open(pkl_ext(filename), 'rb') as f:
        return pickle.load(f)


def datetime_directory(suffix=None, verbose=True):
    now = datetime.now()
    if suffix is None:
        path = 'results/' + now.strftime("%Y%m%d-%H%M%S") + '/'
    else:
        path = 'results/' + now.strftime("%Y%m%d-%H%M%S") + suffix + '/'
    create_dir_at(path, verbose)
    return path


def dict_to_csv(dictionary, filename, indexing='col'):
    filename = csv_ext(filename)
    w = csv.writer(open(filename, 'w'))
    if indexing == 'col':
        w.writerow(dictionary.keys())
        w.writerow(dictionary.values())
    else:
        for key, val in dictionary.items():
            w.writerow([key, val])


def df_to_csv(df, filename):
    filename = csv_ext(filename)
    df.to_csv(filename, index=False, header=True)


def ag_env_directory(exp_dir, agent, env):
    return exp_dir + agent.name + '_' + env.name


def net_directory(env_name):
    return 'results/pretrained/' + env_name


def get_filename(key, exp_dir=None, ag_env_dir=None, net_dir=None, env_name=None, instance_number=None,
                 hidden_layers=None, cst=None):
    """
    Work with unique file names.
    :param key: (str)
    :param exp_dir: (str)
    :param ag_env_dir: (str)
    :param net_dir: (str)
    :param env_name: (str)
    :param instance_number: (int)
    :param hidden_layers: (list)
    :param cst: (float)
    :return: requested filename
    """
    if key == 'global_config':
        return exp_dir + 'global_config'
    elif key == 'env_config':
        return exp_dir + env_name + '_config'
    elif key == 'global_results':
        return ag_env_dir + '/global_results_instance_' + str(instance_number) + '.csv'
    elif key == 'reward_trajectory':
        return ag_env_dir + '/rewards_trajectory_instance_' + str(instance_number) + '.csv'
    elif key == 'q_map':
        return ag_env_dir + '/q_map_instance_' + str(instance_number) + '.csv'
    elif key == 'loss':
        return ag_env_dir + '/loss_instance_' + str(instance_number)
    elif key == 'replay_buffer':
        return ag_env_dir + '/replay_buffer_instance_' + str(instance_number) + '.csv'
    elif key == 'constant_net':
        return net_dir + '/constant_net_' + str(hidden_layers) + '_' + str(round(cst, 1))
    else:
        raise ValueError('Key [' + key + '] not found in get_filename.')


def agents_config_to_md(exp_dir, parameters_list=None, env_name=None):
    filename = exp_dir + 'agents_config.md'
    f = open(filename, 'a+')

    # Environment name before agents config
    if env_name is not None:
        f.write((len(env_name) + 6) * '#' + '\n')
        f.write(3 * ' ' + env_name + '\n')
        f.write((len(env_name) + 6) * '#' + '\n\n')

    # Agents config
    if parameters_list is not None:
        for p in parameters_list:
            max_key_length = max([len(key) for key in p.keys()])
            for key, val in p.items():
                shift = (max_key_length - len(key)) * ' ' + ' : '
                f.write(key + shift + str(val) + '\n')
            f.write('\n')

    f.close()


def benchmark_save(ag_env_dir, agent, env, n_episodes, n_steps, instance_number, trajectory, steps_per_episode,
                   return_per_episode, discounted_return_per_episode, time_per_episode, ag_backup):
    # 1) Global results save
    global_results_df = pandas.DataFrame(columns=['agent', 'environment', 'gamma', 'instance_number', 'episode_number',
                                                  'max_steps', 'performed_steps', 'return', 'discounted_return',
                                                  'time'])
    for i in range(n_episodes):
        global_results_df = global_results_df.append(
            {'agent': agent.name, 'environment': env.name, 'gamma': env.gamma, 'instance_number': instance_number,
             'episode_number': i + 1, 'max_steps': n_steps, 'performed_steps': steps_per_episode[i],
             'return': return_per_episode[i], 'discounted_return': discounted_return_per_episode[i],
             'time': time_per_episode[i]}, ignore_index=True
        )
    global_results_fname = get_filename(key='global_results', ag_env_dir=ag_env_dir, instance_number=instance_number)
    df_to_csv(df=global_results_df, filename=global_results_fname)

    # 2) Reward trajectory save
    trajectory_fname = get_filename(key='reward_trajectory', ag_env_dir=ag_env_dir, instance_number=instance_number)
    df_to_csv(df=trajectory, filename=trajectory_fname)

    # 3) Agent-specific backup
    if ag_backup is not None:
        keys = list(ag_backup.keys())

        # 3.1) Q-table
        if 'q_map' in keys:
            columns = ['x', 'y', 'a', 'q']
            q_map_df = pandas.DataFrame(columns=columns)
            q_map = ag_backup['q_map']
            n_x, n_y, n_a = q_map.shape
            for x in range(1, n_x + 1):
                for y in range(1, n_y + 1):
                    for a in env.actions:
                        q_map_df = q_map_df.append(
                            pandas.DataFrame([[x, y, a, q_map[x - 1][y - 1][a]]], columns=columns))
            q_map_fname = get_filename(key='q_map', ag_env_dir=ag_env_dir, instance_number=instance_number)
            df_to_csv(df=q_map_df, filename=q_map_fname)

        # 3.2) Loss function
        if 'loss' in keys:
            loss_fname = get_filename(key='loss', ag_env_dir=ag_env_dir, instance_number=instance_number)
            save_obj(ag_backup['loss'], filename=loss_fname)

        # 3.2) Replay buffer
        if 'replay_buffer' in keys:
            columns = ['x', 'y', 'frequency']
            replay_buffer_df = pandas.DataFrame(columns=columns)
            for s, freq in ag_backup['replay_buffer'].items():
                replay_buffer_df = replay_buffer_df.append(pandas.DataFrame([[s[0], s[1], freq]], columns=columns))
            replay_buffer_fname = get_filename(key='replay_buffer', ag_env_dir=ag_env_dir,
                                               instance_number=instance_number)
            df_to_csv(df=replay_buffer_df, filename=replay_buffer_fname)



import gym

from drmax.environments.environment import Environment


def make(name, gamma):
    if name == 'CartPole-v0':
        return GymCartPole(name=name, gamma=gamma)
    else:
        raise ValueError('Agent name [' + name + '] not found.')


class GymCartPole(Environment):

    def __init__(self, name, gamma):
        Environment.__init__(self, name=name, actions=[0, 1], gamma=gamma)

        self.env = gym.make('CartPole-v0')

    def get_state_dimension(self):
        return 4

    def get_state_dtype(self):
        return float

    def get_state_magnitude(self):
        return -self.x_threshold, self.x_threshold

    def get_initial_state(self):
        return self.env.reset()

    def step(self, s, a):
        """
        :param s: state
        :param a: action
        :return: r, s_p, is_terminal(s_p)
        """
        s_p, r, is_terminal, _ = self.env.step(a)
        return r, s_p, is_terminal

    def get_info(self):
        """
        Get general information to be saved on disk.
        """
        return {
            'name': self.name,
            'actions': self.actions,
            'gamma': self.gamma
        }



from torch import Tensor
from torch.nn.modules.loss import _Loss
# from torch.nn import functional as F


class OptimisticLoss(_Loss):
    __constants__ = ['reduction']

    def __init__(self, alpha, y_max, optimistic_factor=1.0, size_average=None, reduce=None, reduction: str = 'mean') -> None:
        super(OptimisticLoss, self).__init__(size_average, reduce, reduction)

        self.alpha = alpha
        self.one_m_alpha = 1.0 - self.alpha
        self.y_max = y_max
        self.optimistic_factor = optimistic_factor

    def forward(self, y_prediction: Tensor, y_target: Tensor, y_augmented: Tensor) -> Tensor:
        """
        Forward optimistic loss function
        :param y_prediction: prediction on data set
        :param y_target: target on data set
        :param y_augmented: prediction on additional data for optimism
        :return:
        """
        q_norm = 0.5 / float(y_prediction.size()[0])

        if self.alpha < 1.0:
            u_norm = 1.0 * self.optimistic_factor * self.one_m_alpha / float(y_augmented.size()[0])
            loss = q_norm * (self.alpha * (y_prediction - y_target).pow(2) -
                             self.one_m_alpha * (y_prediction - self.y_max).pow(2)).sum()\
                   + u_norm * (y_augmented - self.y_max).pow(2).sum()
        else:
            return q_norm * (y_prediction - y_target).pow(2).sum()

        return loss


class DualOptimisticLoss(_Loss):
    __constants__ = ['reduction']

    def __init__(self, alpha, y_max, size_average=None, reduce=None, reduction: str = 'mean') -> None:
        super(DualOptimisticLoss, self).__init__(size_average, reduce, reduction)

        self.alpha = alpha
        self.one_m_alpha = 1.0 - self.alpha
        self.y_max = y_max

    def forward(self, y_prediction: Tensor, y_target: Tensor, y_augmented: Tensor) -> Tensor:
        """
        Forward contour-optimistic loss function
        :param y_prediction: prediction on data set
        :param y_target: target on data set
        :param y_augmented: prediction on additional data for optimism
        :return:
        """
        q_norm = self.alpha * 0.5 / float(y_prediction.size()[0])
        # y_target = 0.5 * (y_target + self.y_max)  # Efficient include of symmetric

        if self.alpha < 1.0:
            c_norm = self.one_m_alpha * 0.5 / float(y_augmented.size()[0])
            return q_norm * (y_prediction - y_target).pow(2).sum() + c_norm * (y_augmented - self.y_max).pow(2).sum()
        else:
            return q_norm * (y_prediction - y_target).pow(2).sum()



import timeit
import pandas
import itertools
import numpy as np
import scipy.stats
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm

import drmax.utils.save as save
from drmax.utils.configreader import expand
from drmax.environments.gridworld.gridworld import action_int_to_str


COLORS = [
    'orange',
    'lightblue',
    'peru',
    'tomato',
    'grey',
    'khaki',
    'mediumpurple',
    'pink',
    'yellowgreen',
    'aquamarine'
]
MARKERS = ('o', '^', 's', 'D', 'P', '*')


def rows_to_confidence_intervals_dict(v, confidence=0.9, remove_nan=True):
    n_row = v.shape[0]
    mci = {'mean': [], 'lower_bound': [], 'upper_bound': []}
    for i in range(n_row):
        v_i = v[i][~np.isnan(v[i])] if remove_nan else v[i]
        m, m_lo, m_up = mean_confidence_interval(v_i, confidence=confidence)
        mci['mean'].append(m)
        mci['lower_bound'].append(m_lo)
        mci['upper_bound'].append(m_up)
    return mci


def moving_average(v, window_size):
    v_series = pandas.Series(v)
    windows = v_series.rolling(window_size)
    moving_averages = windows.mean().to_list()
    return moving_averages[window_size - 1:]


def mean_confidence_interval(data, confidence=0.9):
    """
    Compute the mean and confidence interval of the the input data array-like.
    :param data: (array-like)
    :param confidence: probability of the mean to lie in the interval
    :return: (tuple) mean, interval upper-endpoint, interval lower-endpoint
    """
    a = 1.0 * np.array(data)
    if len(a) == 1:
        return a[0], a[0], a[0]
    m, se = np.mean(a), scipy.stats.sem(a)
    if confidence is None:
        return m, None, None
    else:
        h = se * scipy.stats.t.ppf((1.0 + confidence) / 2.0, len(a) - 1)
        return m, m-h, m+h


def plot_trajectories(dictionary, filename, x_key, x_label, y_label, markers_freq, window_size, do_show=False):
    """
    Plot results for all agents
    :param dictionary: (dict) dictionary with the following structure (Example) :
    dictionary = {
        'RMax': {
            x_key: [5, 10, 15, 20, 25, 30]
            'mean': [0, 0, 1, 1, 0, 1]
            'lower_bound': [0, 0, 0, 0, 0, 1]
            'upper_bound': [0, 0, 2, 2, 0, 1]
        },
        'Random': {
            x_key: [5, 10, 15, 20]
            'mean': [0, 0, 1, 0]
            'lower_bound': [0, 0, -2, 0]
            'upper_bound': [0, 0, 3, 0]
        }
    }
    :param filename: (str)
    :param x_label: (str)
    :param y_label: (str)
    :param markers_freq: (int) markers frequency on graphs
    :param window_size: (int) size of the moving average window
    :param do_show: (bool)
    :return:
    """
    fig, ax = plt.subplots()
    ax.set_prop_cycle(color=COLORS)
    m = itertools.cycle(MARKERS)
    for key, val in dictionary.items():
        x, y, y_lo, y_up = val[x_key], val['mean'], val['lower_bound'], val['upper_bound']
        # Compute moving average if necessary
        if window_size > 1:
            if window_size > len(x):
                print('WARNING: required window size in moving average exceeds data length (window_size = ' +
                      str(window_size) + ' data length = ' + str(len(x)) + '). Graph could not be generated.')
                return None
            x, y, y_lo, y_up = (moving_average(v, window_size=window_size) for v in (x, y, y_lo, y_up))
        markers_freq = max(int(len(x) / 10), 1) if markers_freq is None else markers_freq
        ax.plot(x, y, label=key, marker=next(m), markevery=markers_freq)
        ax.fill_between(x, y_lo, y_up, alpha=0.25)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    ax.legend()
    # ax.set_title(title, loc='center')
    ax.grid()
    if do_show:
        plt.show()
    plt.savefig(filename)
    plt.close('all')


def densities_from_df(df, n_instances, n_x, n_y):
    densities = {}
    for i in range(1, n_instances + 1):
        m = np.zeros(shape=(n_x, n_y))
        dfi = df.loc[df['instance_number'] == i]
        for _, row in dfi.iterrows():
            m[row['x'] - 1, row['y'] - 1] = row['frequency']
        densities[i] = np.rot90(m)
    v_min = df["frequency"].min()
    v_max = df["frequency"].max()
    return densities, v_min, v_max


def plot_losses(loss_per_instance_dict, n_instances, filename, env_name, ag_name):
    fig, axes = plt.subplots(n_instances, 1, figsize=(8, 2 * n_instances))
    if n_instances == 1:
        axes = [axes]
    for i in range(n_instances):
        n_losses = len(loss_per_instance_dict[i + 1])
        axes[i].plot(loss_per_instance_dict[i + 1], c='lightsteelblue')
        if n_losses > 10:
            w = int(n_losses / 10)
            x_ma = np.arange(w / 2, n_losses - w / 2 + 1)
            y_ma = moving_average(loss_per_instance_dict[i + 1], window_size=w)
            axes[i].plot(x_ma, y_ma, c='mediumorchid')
        axes[i].set_title('Instance ' + str(i + 1), loc='center')
    fig.suptitle('Losses: ' + ag_name + ' in ' + env_name)
    plt.subplots_adjust(top=0.85, hspace=0.4)
    plt.savefig(filename, bbox_inches='tight')
    plt.close('all')


def plot_densities(densities, v_min, v_max, n_instances, n_x, n_y, filename, env_name, ag_name, cmap='binary'):  # YlGn
    fig, axes = plt.subplots(n_instances, 2, figsize=(2 * n_x, 3 * n_x * n_instances / 4))
    if n_instances == 1:
        axes = [axes]
    for i in range(n_instances):
        im = axes[i][0].imshow(densities[i + 1], cmap=cmap, vmin=0, vmax=v_max)
        plt.colorbar(im, ax=axes[i][0])
        im = axes[i][1].imshow(densities[i + 1], cmap=cmap, norm=LogNorm(vmin=v_min, vmax=v_max))
        plt.colorbar(im, ax=axes[i][1])
        # Set axis for both standard and log-normalized graphs
        for j in [0, 1]:
            axes[i][j].set_title('Instance ' + str(i + 1), loc='center')
            axes[i][j].set_xticks(np.arange(n_x))
            axes[i][j].set_yticks(np.arange(n_y))
            axes[i][j].set_xticklabels(np.arange(1, n_x + 1))
            axes[i][j].set_yticklabels(np.arange(1, n_y + 1))
    fig.suptitle('Replay buffer : ' + ag_name + ' in ' + env_name + '\n(left = raw and right = log-normalized)')
    plt.savefig(filename, bbox_inches='tight')
    plt.close('all')


def plot_q_table(q, n_instances, n_x, n_y, n_a, filename, env_name, ag_name, compute_mean=True, confidence=0.9):
    n_rows = n_instances + 1 if compute_mean else n_instances
    fig, axes = plt.subplots(n_rows, 4, figsize=(2.5 * n_x, 3 * n_x * n_instances / 4))

    # Per-instance Q-table
    for i in range(1, n_instances + 1):
        qi = q.loc[q['instance_number'] == i]
        v_min, v_max = qi['q'].min(), qi['q'].max()
        for a in range(n_a):
            qia = qi.loc[qi['a'] == a][['x', 'y', 'q']].to_numpy()
            m = np.zeros(shape=(n_x, n_y))
            for x, y, q_value in qia:
                m[int(x) - 1][int(y) - 1] = q_value
            m = np.rot90(m)
            im = axes[i - 1, a].imshow(m, cmap='magma', vmin=v_min, vmax=v_max, aspect='auto')
            axes[i - 1, a].set_xticks(np.arange(n_x))
            axes[i - 1, a].set_yticks(np.arange(n_y))
            axes[i - 1, a].set_xticklabels(np.arange(1, n_x + 1))
            axes[i - 1, a].set_yticklabels(np.arange(1, n_y + 1)[::-1])
            axes[i - 1, a].set_title('Q(s, ' + action_int_to_str(a) + ')', loc='center')
        plt.colorbar(im, ax=axes[i - 1, :])

    # Compute mean Q-table
    if compute_mean:
        m, u = {}, {}
        for a in range(n_a):
            qa = q.loc[q['a'] == a]
            m[a], u[a] = np.zeros(shape=(n_x, n_y)), np.zeros(shape=(n_x, n_y))
            for x in range(1, n_x + 1):
                for y in range(1, n_y + 1):
                    q_xya = qa.loc[(qa['x'] == x) & (qa['y'] == y)]['q'].to_numpy()
                    q_mean, _, q_up = mean_confidence_interval(q_xya, confidence=confidence)
                    m[a][x - 1][y - 1], u[a][x - 1][y - 1] = q_mean, q_up - q_mean
            m[a], u[a] = np.rot90(m[a]), np.rot90(u[a])
        v_min, v_max = np.min([m[a] for a in range(n_a)]), np.max([m[a] for a in range(n_a)])
        threshold = v_min + 0.8 * (v_max - v_min)

        # Plot mean Q-table
        for a in range(n_a):
            im = axes[n_rows - 1, a].imshow(m[a], cmap='magma', vmin=v_min, vmax=v_max, aspect='auto')
            axes[n_rows - 1, a].set_xticks(np.arange(n_x))
            axes[n_rows - 1, a].set_yticks(np.arange(n_y))
            axes[n_rows - 1, a].set_xticklabels(np.arange(1, n_x + 1))
            axes[n_rows - 1, a].set_yticklabels(np.arange(1, n_y + 1)[::-1])
            axes[n_rows - 1, a].set_title('Q(s, ' + action_int_to_str(a) + ')', loc='center')
            for x in range(n_x):
                for y in range(n_y):
                    tx = str(round(m[a][x, y], 2)) + '\n±' + str(round(u[a][x, y], 1))
                    cl = ["white", "black"][int(m[a][x, y] > threshold)]
                    axes[n_rows - 1, a].text(y, x, tx, ha="center", va="center", color=cl)
        plt.colorbar(im, ax=axes[n_rows - 1, :])

    # plt.subplots_adjust(right=0.77, wspace=0.25, hspace=0.5)
    fig.suptitle('Q-tables: ' + ag_name + ' in ' + env_name)
    plt.savefig(filename, bbox_inches='tight')
    plt.close('all')


def row_pad(v, length, value=np.nan):
    w = np.full(shape=length - len(v), fill_value=value)
    return np.concatenate((v, w), axis=0)


def retrieve_trajectories(n_instances, ag_env_dir, x_key, file_key):
    trajectories = []
    x = []
    for instance_number in range(1, n_instances + 1):
        filename = save.get_filename(key=file_key, ag_env_dir=ag_env_dir, instance_number=instance_number)
        trajectory_i = pandas.read_csv(filename)
        trajectories.append(trajectory_i)
        if len(x) < len(trajectory_i[x_key]):
            x = trajectory_i[x_key]
    return x, trajectories


def benchmark_plot(exp_dir, agents, environment, markers_freq, reward_plot_ma, return_plot_ma, do_show=False):
    config_filename = save.get_filename(key='global_config', exp_dir=exp_dir)
    config = save.load_obj(config_filename)
    n_env = len(config['environments'])
    matching_env_names = [int(ep['parameters']['name'] == environment.name) for ep in config['environments']]
    n_instances_index = np.dot(matching_env_names, np.arange(n_env))
    n_instances = expand(config['n_instances'], n_env)[n_instances_index]

    # 1) Reward trajectories

    # Buffers to store trajectories
    rewards = {}
    cumulative_rewards = {}
    discounted_rewards = {}
    cumulative_discounted_rewards = {}
    return_per_episode = {}
    discounted_return_per_episode = {}

    for ag in agents:
        ag_env_dir = save.ag_env_directory(exp_dir=exp_dir, agent=ag, env=environment)

        steps, reward_trajectories = retrieve_trajectories(n_instances, ag_env_dir, x_key='step',
                                                           file_key='reward_trajectory')
        episodes, return_trajectories = retrieve_trajectories(n_instances, ag_env_dir, x_key='episode_number',
                                                              file_key='global_results')

        for col_key, x_key, x, buffer, trajectories in [
            ('r', 'step', steps, rewards, reward_trajectories),
            ('cr', 'step', steps, cumulative_rewards, reward_trajectories),
            ('dr', 'step', steps, discounted_rewards, reward_trajectories),
            ('cdr', 'step', steps, cumulative_discounted_rewards, reward_trajectories),
            ('return', 'episode', episodes, return_per_episode, return_trajectories),
            ('discounted_return', 'episode', episodes, discounted_return_per_episode, return_trajectories),
        ]:
            max_length = max([len(t[col_key]) for t in trajectories])
            v = np.transpose([row_pad(t[col_key].to_numpy(), length=max_length) for t in trajectories])
            ag_traj_dict = rows_to_confidence_intervals_dict(v)  # mean lower upper dictionary
            ag_traj_dict[x_key] = x
            buffer[ag.name] = ag_traj_dict

    for ma in reward_plot_ma:
        for dct, sfx, x_key, xlb, ylb in [
            (rewards, '_rewards_ma' + str(ma) + '.pdf', 'step', 'Decision epoch', 'Rewards'),
            (cumulative_rewards, '_cumulative_rewards_ma' + str(ma) + '.pdf', 'step', 'Decision epoch',
             'Cumulative Rewards'),
            (discounted_rewards, '_discounted_rewards_ma' + str(ma) + '.pdf', 'step', 'Decision epoch',
             'Discounted Rewards'),
            (cumulative_discounted_rewards, '_cumulative_discounted_rewards_ma' + str(ma) + '.pdf', 'step',
             'Decision epoch', 'Cumulative Discounted Rewards')
        ]:
            plot_trajectories(dct, filename=exp_dir + environment.name + sfx, x_key=x_key, x_label=xlb, y_label=ylb,
                              markers_freq=markers_freq, window_size=ma, do_show=do_show)

    for ma in return_plot_ma:
        for dct, sfx, x_key, xlb, ylb in [
            (return_per_episode, '_return_per_episode_ma' + str(ma) + '.pdf', 'episode', 'Episode',
             'Return per episode'),
            (discounted_return_per_episode, '_discounted_return_per_episode_ma' + str(ma) + '.pdf', 'episode',
             'Episode', 'Discounted return per episode')
        ]:
            plot_trajectories(dct, filename=exp_dir + environment.name + sfx, x_key=x_key, x_label=xlb, y_label=ylb,
                              markers_freq=markers_freq, window_size=ma, do_show=do_show)

    # 2) Agent-specific data (if it exists)
    for ag in agents:
        ag_env_dir = save.ag_env_directory(exp_dir=exp_dir, agent=ag, env=environment)

        # 2.1) Q-map
        try:
            q = pandas.DataFrame()
            for instance_number in range(1, n_instances + 1):
                q_map_i_fname = save.get_filename(key='q_map', ag_env_dir=ag_env_dir, instance_number=instance_number)
                q_map_i = pandas.read_csv(q_map_i_fname)
                q_map_i['instance_number'] = q_map_i.shape[0] * [instance_number]
                q = pandas.concat([q, q_map_i])
            n_x, n_y, n_a = max(q['x']), max(q['y']), 4  # max(df['a']) + 1
            # Plot Q-table for each instance + mean Q-table
            fname = exp_dir + environment.name + '_' + ag.name + '_q_table.pdf'
            plot_q_table(q, n_instances, n_x, n_y, n_a, filename=fname, env_name=environment.name, ag_name=ag.name,
                         compute_mean=True, confidence=0.9)
        except IOError:
            pass

        # 2.2) Replay-buffer
        try:
            rb = pandas.DataFrame()
            for instance_number in range(1, n_instances + 1):
                replay_buffer_i_fname = save.get_filename(key='replay_buffer', ag_env_dir=ag_env_dir,
                                                          instance_number=instance_number)
                replay_buffer_i = pandas.read_csv(replay_buffer_i_fname)
                replay_buffer_i['instance_number'] = replay_buffer_i.shape[0] * [instance_number]
                rb = pandas.concat([rb, replay_buffer_i])

            env_config_filename = save.get_filename(key='env_config', exp_dir=exp_dir, env_name=environment.name)
            env_config = save.load_obj(filename=env_config_filename)
            grid_width, grid_height = env_config['width'], env_config['height']

            densities, v_min, v_max = densities_from_df(rb, n_instances, n_x=grid_width, n_y=grid_height)
            fname = exp_dir + environment.name + '_' + ag.name + '_replay_buffer.pdf'
            plot_densities(densities, v_min, v_max, n_instances, n_x=grid_width, n_y=grid_height, filename=fname,
                           env_name=environment.name, ag_name=ag.name)
        except IOError:
            pass

        # 2.3) Loss
        try:
            loss_per_instance_dict = {}
            for instance_number in range(1, n_instances + 1):
                loss_fname = save.get_filename(key='loss', ag_env_dir=ag_env_dir, instance_number=instance_number)
                loss_i = save.load_obj(filename=loss_fname)
                loss_per_instance_dict[instance_number] = loss_i
            fname = exp_dir + environment.name + '_' + ag.name + '_losses.pdf'
            plot_losses(loss_per_instance_dict, n_instances, filename=fname, env_name=environment.name,
                        ag_name=ag.name)
        except IOError:
            pass



"""
Abstract agent class
"""


class Agent(object):

    def __init__(self, name, actions, gamma):
        self.name = name
        self.actions = actions
        self.gamma = gamma

    def act(self, s, r, is_terminal):
        pass

    def end_of_episode(self):
        pass

    def re_init(self):
        """
        Re-initialization for multiple instances.
        :return: None
        """
        self.__init__(name=self.name, actions=self.actions, gamma=self.gamma)

    def get_backup(self):
        return None



import random

from drmax.agents.agent import Agent


class Random(Agent):

    def __init__(self, name, actions, gamma):
        Agent.__init__(self, name=name, actions=actions, gamma=gamma)

    def act(self, s, r, is_terminal):
        return random.choice(self.actions)

    def re_init(self):
        """
        Re-initialization for multiple instances.
        :return: None
        """
        self.__init__(name=self.name, actions=self.actions, gamma=self.gamma)



"""
Abstract environment class
"""


class Environment(object):

    def __init__(self, name, actions, gamma):
        self.name = name
        self.actions = actions
        self.gamma = gamma

    def get_state_dimension(self):
        return None

    def get_state_dtype(self):
        return None

    def get_state_magnitude(self):
        return None

    def get_initial_state(self):
        return None

    def step(self, s, a):
        """
        :param s: state
        :param a: actionSmall 
        :return: r, s_p, is_terminal(s_p)
        """
        return 0.0, None, False

    def get_info(self):
        """
        Get general information to be saved on disk.
        """
        return {
            'name': self.name,
            'actions': self.actions,
            'gamma': self.gamma
        }



import random
from collections import defaultdict

from drmax.agents.agent import Agent


class QLearning(Agent):

    def __init__(self, name, actions, gamma, learning_rate, epsilon):
        Agent.__init__(self, name=name, actions=actions, gamma=gamma)

        self.learning_rate = learning_rate
        self.epsilon = epsilon

        self.q_max = 1.0 / (1.0 - gamma)
        self.Q = defaultdict(lambda: defaultdict(lambda: self.q_max))
        self.prev_s = None
        self.prev_a = None

        self.learns_Q_function = True

    def re_init(self):
        """
        Re-initialization for multiple instances.
        :return: None
        """
        self.__init__(name=self.name, actions=self.actions, gamma=self.gamma, learning_rate=self.learning_rate,
                      epsilon=self.epsilon)

    def reset(self):
        """
        Reset the attributes to initial state (called between instances).
        :return: None
        """
        self.Q = defaultdict(lambda: defaultdict(lambda: self.q_max))
        self.prev_s = None
        self.prev_a = None

    def act(self, s, r, is_terminal):
        """
        Acting method called online during learning.
        :param s: current state of the agent
        :param r: (float) received reward for the previous transition
        :param is_terminal: (bool) True if s is terminal
        :return: return the epsilon-greedy action wrt the current learned Q-function.
        """
        self.update(self.prev_s, self.prev_a, r, s)

        a = self.epsilon_greedy_action(s, self.Q)

        self.prev_a = a
        self.prev_s = s

        return a

    def update(self, s, a, r, s_p):
        """
        Update the current Q-function
        :param s: state
        :param a: action
        :param r: (float) reward
        :param s_p:  next state
        :return: None
        """
        if s is not None and a is not None:
            self.Q[s][a] += self.learning_rate * (r + self.gamma * self.greedy_action(s_p, self.Q)[1] - self.Q[s][a])

    def end_of_episode(self):
        """
        Reset between episodes within the same MDP.
        :return: None
        """
        self.prev_s = None
        self.prev_a = None

    def epsilon_greedy_action(self, s, f):
        if random.random() < self.epsilon:
            return random.choice(self.actions)
        else:
            return self.greedy_action(s, f)[0]

    def greedy_action(self, s, f):
        """
        Compute the greedy action wrt the input function of (s, a).
        :param s: state at which the upper-bound is evaluated
        :param f: input function of (s, a)
        :return: return the greedy action and the achieved value by this action.
        """
        a_star = random.choice(self.actions)
        u_star = f[s][a_star]
        for a in self.actions:
            u_s_a = f[s][a]
            if u_s_a > u_star:
                u_star = u_s_a
                a_star = a
        return a_star, u_star

    def q_forward(self, s, a):
        """
        Evaluate Q function at (s, a)
        :param s: state
        :param a: action
        :return: Q-value of (s, a)
        """
        return self.Q[s][a]
